/* CvsModule.m created by stephane on Wed 08-Sep-1999 */
/*$Id: CvsModule.m,v 1.14 2002/06/14 12:35:17 stephane Exp $*/

// Copyright (c) 1997-2001, Sen:te Ltd.  All rights reserved.
//
// Use of this source code is governed by the license in OpenSourceLicense.html
// found in this distribution and at http://www.sente.ch/software/ ,  where the
// original version of this source code can also be found.
// This notice may not be removed from this file.

#import "CvsModule.h"
#import "../SenFoundationCVLCandidates.subproj/NSString_SenCaseInsensitiveComparison.h"
#ifdef MACOSX
#import <CvsRepository.h>
#import <NSString+Lines.h>
#else
#import <Repositories/CvsRepository.h>
#import <SenFoundation/NSString+Lines.h>
#endif
#import <SenOyster/NSString.SenOysterRegex.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSEnumerator.h>
#import <Foundation/NSException.h>
#import <Foundation/NSString.h>


// Perhaps we should not use either \r\n or \n. Perhaps we should always use both EOL
#ifndef WIN32
static NSString	*CvsModuleLineEndingWithBackslashesBySpaceSubstitutionPattern = @"s/\\\\\\n/ /g"; // Replace trailing \ with space
static NSString	*CvsModuleLineBeginningWithDiezesByVoidSubstitutionPattern = @"s/.*^#.*\\n/\\n/g"; // Remove comment line (MUST begin with #)
//static NSString	*CvsModuleLineBeginningWithSpacesByVoidSubstitutionPattern = @"s/.*^\\s+//g"; // Remove trailing spaces
//static NSString	*CvsModuleLineWithDiezesByVoidSubstitutionPattern = @"s/#.*\\n/\\n/g"; // Remove comments (both inline and full line)
//static NSString	*CvsModuleLineWithSpacesByLinesSubstitutionPattern = @"s/\\s+\\n/\\n/g"; // Remove trailing spaces
#else WIN32
static NSString	*CvsModuleLineEndingWithBackslashesBySpaceSubstitutionPattern = @"s/\\\\\\r\\n/ /g"; // Replace trailing \ with space
static NSString	*CvsModuleLineBeginningWithDiezesByVoidSubstitutionPattern = @"s/.*^#.*\\r\\n/\\r\\n/g"; // Remove comment line (MUST begin with #)
//static NSString	*CvsModuleLineWithSuffixBackslashesBySpaceSubstitutionPattern = @"s/\\\\\\r\\n/ /g"; // Replace trailing \ with space
//static NSString	*CvsModuleLineWithSpacesByLinesSubstitutionPattern = @"s/\\s+\\r\\n/\\r\\n/g";
//static NSString	*CvsModuleLineWithDiezesByVoidSubstitutionPattern = @"s/\s*#.*\\n//g"; // Remove comment lines
#endif
static NSString	*CvsModuleLineContainingDiezeByVoidSubstitutionPattern = @"s/#.*//g"; // Remove inline comment

@implementation CvsModule

// Module list can be retrieved with cvs checkout -c and cvs checkout -s commands
// Using -c option, we get a parsed module list with all options; all components are separated by a space
// Description of one module can extend beyond one line, beginning with spaces on following lines.
// Inline comments are not retrieved.
// Using -s option, we can get inline comments, but not the options!
// So, to get all info, we'd need to perform two requests, and merge their results
// To save a new module list, we could simply replace module definitions one after the other in the original file,
// thus we don't loose general comments and file structure.
// Using cvs built-in functionalities will avoid us to make errors in modules parsing...

+ (NSArray *) parseModuleDescription:(NSString *)moduleDescription forRepository:(CvsRepository *)aRepository generatesModules:(BOOL)generates
// moduleDescription is the raw content of file modules
// If generates is YES, will return an array of CvsModules or raise an exception if an error occurred during parsing
// If generates is NO, will return an array of error messages if errors occurred during parsing, or nil
{
    NSMutableString		*buffer = [moduleDescription mutableCopy];
    NSArray				*linesArray;
    int					num = 0;
    NSMutableDictionary	*modules = nil;
    NSMutableArray		*errors = nil;

    // <modules> parsing is defined in project cvs: src/mkmodules.c:write_dbmfile()
    // 1) Check for trailing backslashes and replace them and the following EOL with a space
    // 2) Check for lines beginning with # (MUST be first line character), and remove them
    // 3) Remove prefixing spaces (isspace())
    // 4) Remove empty lines
    // 5) Check each key is unique and has a value
    // Module descriptions are parsed later in src/modules.c:do_module()
    // cvs uses getopt() with ' ' and '\t' as authorized separators
    // In values, module names beginning with ! are ignored (for aliases only)
    // Everything following # is ignored
    // When parsing options, if an option appears more than once, the last occurence is used
    // Authorized options are -a (with one and more arguments)
    // or -i, -o, -e, -t, -u, -d (with one argument)
    // or -l (without any argument)
    // Option -s (with one argument) is ignored
    // cvs uses GNU's getopt(); getopt() is defined in <bsd/unistd.h>

    // 1) Check for lines extended on successive lines with \
    num = [buffer applyReplacementOperator:CvsModuleLineEndingWithBackslashesBySpaceSubstitutionPattern];
    // 2) Remove comment lines
    num = [buffer applyReplacementOperator:CvsModuleLineBeginningWithDiezesByVoidSubstitutionPattern];
    // Remove trailing spaces
//    num = [buffer applyReplacementOperator:CvsModuleLineWithSpacesByLinesSubstitutionPattern];

    linesArray = [buffer lines];
    num = [linesArray count];
    if(num){
        NSString		*lineString;
        NSEnumerator	*lineEnum = [linesArray objectEnumerator];

        errors = [NSMutableArray array];
        modules = [NSMutableDictionary dictionaryWithCapacity:num];
        while(lineString = [lineEnum nextObject]){
            CvsModule		*aModule;
            NSArray			*wordsFromLine;
            int				wordCount;
            NSEnumerator	*wordEnum;
            NSString		*aWord;

            lineString = [lineString stringByApplyingReplacementOperator:CvsModuleLineContainingDiezeByVoidSubstitutionPattern]; // Remove inline comment
            wordsFromLine = [lineString componentsSeparatedByOperator:@"/[ \\t]+/"]; // Authorized separators are space and tab
            wordCount = [wordsFromLine count];
            if(wordCount == 0)
                continue; // It file has an empty line, it will be put into linesArray, so let's get rid of it

            if(wordCount == 1){
                [errors addObject:[NSString stringWithFormat:@"Module %@ is wrongly defined: contains only a symbolic name without any content.\n%@", [wordsFromLine lastObject], lineString]];
                continue;
            }

            aModule = [[self allocWithZone:[aRepository zone]] init];
            wordEnum = [wordsFromLine objectEnumerator];

            [aModule setRepository:aRepository];
            [aModule setSymbolicName:[wordEnum nextObject]]; // Index 0
            aWord = [wordEnum nextObject]; // Index 1
            if([aWord isEqualToString:@"-a"]){
                // Option MUST be separated from file list; cvs does not like it otherwise
                if(wordCount < 3){
                    [errors addObject:[NSString stringWithFormat:@"Module %@ is wrongly defined: no aliases but option -a.\n%@", [aModule symbolicName], lineString]];
                    continue;
                }
                // Alias names beginning with ! will be ignored by cvs
                [aModule setAliases:[wordsFromLine subarrayWithRange:NSMakeRange(2, wordCount - 2)]];
            }
            else{
                NSMutableDictionary	*theOptions = [NSMutableDictionary dictionaryWithCapacity:10];
                NSMutableArray		*theFiles = [NSMutableArray array];
                NSMutableArray		*additionalModulesNames = [NSMutableArray array];

                do{
                    if([aWord hasPrefix:@"-"]){
#warning (Stephane) BUG 1000097 - Some unsupported option definition types!
						// WARNING: options may be appended in a single block, like in -ld,
						// and option arguments may be appended to option, like in -dMyDirectory
						// Currently we do NOT support this!!!
                        NSString	*param = [wordEnum nextObject];

                        if([aModule directory] != nil){
                            [errors addObject:[NSString stringWithFormat:@"Module %@ is wrongly defined: option %@ defined after directory.\n%@", [aModule symbolicName], aWord, lineString]];
                            continue;
                        }
                        if([aWord isEqualToString:@"-l"])
                            param = @""; // Option -l requires no parameter! Top-level directory only -- do not recurse
                        if(param == nil){
                            [errors addObject:[NSString stringWithFormat:@"Module %@ is wrongly defined: no param after option %@.\n%@", [aModule symbolicName], aWord, lineString]];
                            continue;
                        }
                        [theOptions setObject:param forKey:aWord]; // We don't care if option was already defined, like cvs
                    }
                    else{
                        // WARNING: documentation on cvs has a bug: in 1.10.3, they allow to define a module like this:
                        // myModule &anotherModule
                        // without any directory...
                        if(![aWord hasPrefix:@"&"] && ![aModule directory])
                            [aModule setDirectory:aWord];
                        else if([aWord hasPrefix:@"&"]){
                            if([aWord length] <= 1){
                                [errors addObject:[NSString stringWithFormat:@"Module %@ is wrongly defined: no module name after &.\n%@", [aModule symbolicName], lineString]];
                                continue;
                            }
                            [additionalModulesNames addObject:[aWord substringFromIndex:1]];
                        }
                        else{
/*
 							// cvs supports this, but docs seems to tell the opposite
 							// Let's support is anyway
                            if([additionalModulesNames count] != 0){
                                [errors addObject:[NSString stringWithFormat:@"Module %@ is wrongly defined: adds file definition after modules enumeration.\n%@", [aModule symbolicName], lineString]];
                                continue;
                            }
 */
                            [theFiles addObject:aWord];
                        }
                    }
                }while(aWord = [wordEnum nextObject]);

                if(![aModule directory] && ![theFiles count] && ![additionalModulesNames count])
                    [errors addObject:[NSString stringWithFormat:@"Module %@ is wrongly defined: it has no directory, not even a module or file\n%@", [aModule symbolicName], lineString]];
                else{
                    if([theFiles count])
                        [aModule setFiles:theFiles];
                    if([additionalModulesNames count])
                        [aModule setAdditionalModuleNames:additionalModulesNames];
                    if([theOptions count])
                        [aModule setOptions:theOptions];
                }
            }
            
            if([modules objectForKey:[aModule symbolicName]])
                [errors addObject:[NSString stringWithFormat:@"Module %@ is defined more than once -- new definition ignored", [aModule symbolicName]]];
            else
                [modules setObject:aModule forKey:[aModule symbolicName]];

            [aModule release];
        }
    }

    [buffer release];

    if(errors && [errors count])
        if(generates){
            [[NSException exceptionWithName:NSInvalidArgumentException reason:@"Error during modules parsing" userInfo:[NSDictionary dictionaryWithObject:errors forKey:@"Errors"]] raise];
            return nil; // Never reached, but compiler will be happy with this ;-)
        }
        else
            return errors;

    if(generates)
        return [modules allValues];
    else
        return nil;
}

+ (NSArray *) modulesWithContentsOfFile:(NSString *)filename forRepository:(CvsRepository *)aRepository
{
    volatile NSArray	*result = nil;
    
    NS_DURING
        result = [self parseModuleDescription:[NSString stringWithContentsOfFile:filename] forRepository:aRepository generatesModules:YES];
    NS_HANDLER
        NSString	*errors = [[[localException userInfo] objectForKey:@"Errors"] componentsJoinedByString:@"\n"];

        // Ideally we should log the errors in our Console, shouldn't we? But we shouldn't depend on the console in this code...
        NSLog(@"%@: %@%@", [localException name], [localException reason], errors ? [@"\n" stringByAppendingString:errors]:@"");
    NS_ENDHANDLER

    return (NSArray *)result;
}

- (void) dealloc
{
    [symbolicName release];
    [aliases release];
    [directory release];
    [files release];
    [additionalModuleNames release];
    [options release];
    
    [super dealloc];
}

- (NSString *) description
{
    NSMutableString	*description = [symbolicName mutableCopy];

    if(aliases && [aliases count]){
        [description appendString:@" -a "];
        [description appendString:[aliases componentsJoinedByString:@" "]];
    }
    else{
        if(options && [options count]){
            NSEnumerator	*keyEnum = [options keyEnumerator];
            NSString		*aKey;

            while(aKey = [keyEnum nextObject])
                [description appendFormat:@" %@ %@", aKey, [options objectForKey:aKey]];
        }
        [description appendString:@" "];
        [description appendString:directory];
        if(files && [files count]){
            [description appendString:@" "];
            [description appendString:[files componentsJoinedByString:@" "]];
        }
        if(additionalModuleNames && [additionalModuleNames count]){
            [description appendString:@" &"];
            [description appendString:[additionalModuleNames componentsJoinedByString:@" &"]];
        }        
    }

    if(repository)
        [description appendFormat:@" (in %@)", [repository root]];

    return [description autorelease];
}

- (CvsRepository *) repository
{
    return repository;
}

- (void) setRepository:(CvsRepository *)aRepository
{
    repository = aRepository;
}

- (NSString *) symbolicName
{
    return symbolicName;
}

- (void) setSymbolicName:(NSString *)aSymbolicName
{
    [symbolicName autorelease];
    symbolicName = [aSymbolicName retain];
}

- (NSArray *) aliases
{
    return aliases;
}

- (void) setAliases:(NSArray *)names
{
    [aliases autorelease];
    aliases = [names retain];
}

- (NSString *) directory
{
    return directory;
}

- (void) setDirectory:(NSString *)aDirectory
{
    [directory autorelease];
    directory = [aDirectory retain];
}

- (NSArray *) files
{
    return files;
}

- (void) setFiles:(NSArray *)names
{
    [files autorelease];
    files = [names retain];
}

- (NSArray *) additionalModuleNames
{
    return additionalModuleNames;
}

- (void) setAdditionalModuleNames:(NSArray *)names
{
    [additionalModuleNames autorelease];
    additionalModuleNames = [names retain];
}

- (NSString *) outputDirectoryName
{
    NSString	*aName;

    if(options && (aName = [options objectForKey:@"-d"]))
        return aName;
    
    // WARNING: one case is not handled at all: aliases!
    // If a module is defined by aliasing some paths/modules, many workareas can be opened at once
    // In case of aliases, cvs will create directories as defined by aliases, and not as defined by
    // the module symbolicName.

    if(aliases && [aliases count]){
        // Currently, let's at least support 1 alias item
        CvsModule	*aliasedModule = [[self repository] moduleWithSymbolicName:[aliases objectAtIndex:0]];

        if(aliasedModule)
            return [aliasedModule outputDirectoryName];
        else
            // By path
            return [aliases objectAtIndex:0];
        // If we want to support multiple alias items, we should change -[CVLDelegate requestCompleted:]
        // and -[CvsCheckoutRequest destinationPath] to get all the aliases
    }
    
    return symbolicName;
}

- (void) setOutputDirectoryName:(NSString *)anOutputDirectoryName
{
    if(!options)
        options = [NSMutableDictionary dictionaryWithObject:anOutputDirectoryName forKey:@"-d"];
    else
        [options setObject:anOutputDirectoryName forKey:@"-d"];
}

- (NSString *) status
{
    NSString	*aName = nil;
    
    if(options)
        aName = [options objectForKey:@"-s"];
    return aName;
}

- (void) setStatus:(NSString *)aStatus
{
    if(!options)
        options = [NSMutableDictionary dictionaryWithObject:aStatus forKey:@"-s"];
    else
        [options setObject:aStatus forKey:@"-s"];
}

- (NSDictionary *) options
{
    return options;
}

- (void) setOptions:(NSDictionary *)someOptions
{
    [options autorelease];
    options = [someOptions mutableCopyWithZone:[self zone]];
}

- (NSComparisonResult) compareSymbolicName:(CvsModule *)aModule
{
    return [symbolicName senCaseInsensitiveCompare:[aModule symbolicName]];
}

@end
